import os

from unittest import mock

from aioresponses import aioresponses
import asyncpg
import pytest

from minicli import run

import udata_hydra.cli  # noqa - this register the cli cmds
from udata_hydra.crawl import insert_check

DATABASE_URL = os.getenv("DATABASE_URL", "postgresql://postgres:postgres@localhost:5433/postgres")
pytestmark = pytest.mark.asyncio


# this really really really should run first (or "prod" db will get erased)
@pytest.fixture(autouse=True, scope="session")
def setup():
    with mock.patch.dict(os.environ, {"DATABASE_URL": DATABASE_URL}):
        yield


@pytest.fixture(autouse=True)
async def mock_pool(mocker):
    """This avoids having different pools attached to different event loops"""
    m = mocker.patch("udata_hydra.context.pool")
    pool = await asyncpg.create_pool(dsn=DATABASE_URL, max_size=50)
    m.return_value = pool


@pytest.fixture
def catalog_content():
    with open("tests/catalog.csv", "rb") as cfile:
        return cfile.read()


@pytest.fixture
def setup_catalog(catalog_content, rmock):
    catalog = "https://example.com/catalog"
    rmock.get(catalog, status=200, body=catalog_content)
    run("init_db", drop=True, table=None, index=True, reindex=False)
    run("load_catalog", url=catalog)


@pytest.fixture
def produce_mock(mocker):
    mocker.patch("udata_hydra.crawl.produce")
    mocker.patch("udata_hydra.datalake_service.produce")

@pytest.fixture
def rmock():
    # passthrough for local requests (aiohttp TestServer)
    with aioresponses(passthrough=["http://127.0.0.1"]) as m:
        yield m


@pytest.fixture
async def db():
    conn = await asyncpg.connect(dsn=DATABASE_URL)
    yield conn
    await conn.close()


@pytest.fixture
async def fake_check(db):
    async def _fake_check(
        status=200,
        error=None,
        timeout=False,
        resource=1,
        created_at=None,
        headers={"x-do": "you"},
    ):
        data = {
            "url": f"https://example.com/resource-{resource}",
            "domain": "example.com",
            "status": status,
            "headers": headers,
            "timeout": timeout,
            "response_time": 0.1,
            "error": error,
        }
        id = await insert_check(data)
        if created_at:
            data["created_at"] = created_at
            await db.execute(
                """
                UPDATE checks
                SET created_at = $1
                WHERE id = $2
            """,
                created_at,
                id,
            )
        return data

    return _fake_check



import json

from kafka import KafkaProducer

DOCUMENT_LOURD = {
    "id": "5b7ffc618b4c4169d30727e0",
    "title": "Base Sirene des entreprises et de leurs établissements (SIREN, SIRET)",
    "description": "satisfaite.",
    "acronym": None,
    "url": "http://dev-02.data.gouv.fr/fr/datasets/base-sirene-des-entreprises-et-de-leurs-etablissements-siren-siret/",
    "tags": [
        "association",
        "companies",
        "entreprises",
        "etablissements",
        "register",
        "registre",
        "repertoire",
        "siren",
        "sirene",
        "siret",
    ],
    "license": "lov2",
    "badges": ["spd"],
    "frequency": "monthly",
    "created_at": "2018-08-24T14:38:57",
    "views": 29530,
    "followers": 56,
    "reuses": 19,
    "featured": 0,
    "resources_count": 15,
    "organization": {
        "id": "534fff81a3a7292c64a77e5c",
        "name": "Institut National de la Statistique et des Etudes Economiques (Insee)",
        "public_service": 1,
        "followers": 615,
    },
    "owner": None,
    "format": [
        "zip",
        "zip",
        "zip",
        "zip",
        "zip",
        "pdf",
        "pdf",
        "pdf",
        "pdf",
        "pdf",
        "csv",
        "csv",
        "csv",
        "csv",
        "csv",
    ],
    "schema": [],
    "extras": {},
    "resources": [
        {
            "id": "88fbb6b4-0320-443e-b739-b4376a012c32",
            "url": "https://files.data.gouv.fr/insee-sirene/StockEtablissementHistorique_utf8.zip",
            "format": "zip",
            "title": "Sirene : Fichier StockEtablissementHistorique du 01 Mars 2022",
            "schema": {},
            "latest": "http://dev-02.data.gouv.fr/fr/datasets/r/88fbb6b4-0320-443e-b739-b4376a012c32",
            "description": None,
            "filetype": "remote",
            "type": "main",
            "created_at": "2021-09-01T07:46:43",
            "modified": "2022-03-01T22:57:39",
            "published": "2022-03-01T21:57:39",
            "extras": {
                "check:status": 204,
                "check:available": True,
                "check:date": "2022-01-24T15:28:58",
                "check:count-availability": 3,
            },
        },
        {
            "id": "0651fb76-bcf3-4f6a-a38d-bc04fa708576",
            "url": "https://files.data.gouv.fr/insee-sirene/StockEtablissement_utf8.zip",
            "format": "zip",
            "title": "Sirene : Fichier StockEtablissement du 01 Mars 2022",
            "schema": {},
            "latest": "http://dev-02.data.gouv.fr/fr/datasets/r/0651fb76-bcf3-4f6a-a38d-bc04fa708576",
            "description": None,
            "filetype": "remote",
            "type": "main",
            "created_at": "2021-09-01T07:39:48",
            "modified": "2022-03-01T22:57:39",
            "published": "2022-03-01T21:57:39",
            "extras": {},
        },
        {
            "id": "0835cd60-2c2a-497b-bc64-404de704ce89",
            "url": "https://files.data.gouv.fr/insee-sirene/StockUniteLegaleHistorique_utf8.zip",
            "format": "zip",
            "title": "Sirene : Fichier StockUniteLegaleHistorique du 01 Mars 2022",
            "schema": {},
            "latest": "http://dev-02.data.gouv.fr/fr/datasets/r/0835cd60-2c2a-497b-bc64-404de704ce89",
            "description": None,
            "filetype": "remote",
            "type": "main",
            "created_at": "2021-09-01T07:21:25",
            "modified": "2022-03-01T22:57:39",
            "published": "2022-03-01T21:57:38",
            "extras": {},
        },
        {
            "id": "9c4d5d9c-4bbb-4b9c-837a-6155cb589e26",
            "url": "https://files.data.gouv.fr/insee-sirene/StockEtablissementLiensSuccession_utf8.zip",
            "format": "zip",
            "title": "Sirene : Fichier StockEtablissementLiensSuccession du 01 Mars 2022",
            "schema": {},
            "latest": "http://dev-02.data.gouv.fr/fr/datasets/r/9c4d5d9c-4bbb-4b9c-837a-6155cb589e26",
            "description": None,
            "filetype": "remote",
            "type": "main",
            "created_at": "2021-09-01T05:35:10",
            "modified": "2022-03-01T22:57:39",
            "published": "2022-03-01T21:57:39",
            "extras": {},
        },
        {
            "id": "825f4199-cadd-486c-ac46-a65a8ea1a047",
            "url": "https://files.data.gouv.fr/insee-sirene/StockUniteLegale_utf8.zip",
            "format": "zip",
            "title": "Sirene : Fichier StockUniteLegale du 01 Mars 2022",
            "schema": {},
            "latest": "http://dev-02.data.gouv.fr/fr/datasets/r/825f4199-cadd-486c-ac46-a65a8ea1a047",
            "description": None,
            "filetype": "remote",
            "type": "main",
            "created_at": "2021-09-01T05:34:07",
            "modified": "2022-03-01T22:57:39",
            "published": "2022-03-01T21:57:39",
            "extras": {},
        }
    ],
    "geozones": [
        {
            "id": "country:fr",
            "name": "France",
            "keys": ["fr", "fra", "250", "fr", "3017382", "fr", "2202162"],
        },
        {"id": "country-group:ue"},
        {"id": "country-group:world"},
    ],
    "granularity": "other",
}


DOCUMENT_LEGER = {
    "id": "6065c5cd9f89bfd828bcbf5c",
    "title": "Nombre de personnes rickrollées sur data.gouv.fr",
    "description": "A l'occasion du 1er avril, data.gouv.fr s'est donné pour mission de [Rickroll](https://fr.wikipedia.org/wiki/Rickroll) un nombre important d'usagers. Dans un souci de transparence, les indicateurs relatifs à cette mission sont publiés ici en open data sous forme de fichier CSV.\n\nLes données sont issues de [stats.data.gouv.fr](https://stats.data.gouv.fr/index.php?module=CoreHome&action=index&idSite=109&period=range&date=previous30#?idSite=109&period=day&date=2021-04-01&segment=&category=General_Actions&subcategory=General_Outlinks).",
    "acronym": None,
    "url": "http://dev-02.data.gouv.fr/fr/datasets/nombre-de-personnes-rickrollees-sur-data-gouv-fr/",
    "tags": ["never-gonna-give-you-up"],
    "license": "lov2",
    "badges": [],
    "frequency": "punctual",
    "created_at": "2021-04-01T15:08:29",
    "views": 160,
    "followers": 2,
    "reuses": 0,
    "featured": 0,
    "resources_count": 1,
    "organization": {
        "id": "534fff75a3a7292c64a77de4",
        "name": "Etalab",
        "public_service": 1,
        "followers": 356,
    },
    "owner": None,
    "format": ["csv"],
    "schema": [],
    "extras": {},
    "resources": [
        {
            "id": "503c39b0-58d8-45fe-ace0-55795bb0a6b7",
            "url": "https://static.data.gouv.fr/resources/nombre-de-personnes-rickrollees-sur-data-gouv-fr/20210401-182031/rickroll.csv",
            "format": "csv",
            "title": "rickroll.csv",
            "schema": {},
            "latest": "http://dev-02.data.gouv.fr/fr/datasets/r/503c39b0-58d8-45fe-ace0-55795bb0a6b7",
            "description": None,
            "filetype": "file",
            "type": "main",
            "created_at": "2021-04-01T15:10:12",
            "modified": "2021-04-01T18:20:33",
            "published": "2021-04-01T15:10:12",
            "extras": {},
        }
    ],
    "temporal_coverage_start": "2021-04-01T00:00:00",
    "temporal_coverage_end": "2021-04-02T00:00:00",
}


def main():
    producer = KafkaProducer(
        bootstrap_servers="localhost:9092",
        value_serializer=lambda v: json.dumps(v).encode("utf-8"),
    )
    key = "5b7ffc618b4c4169d30727e0".encode("utf-8")

    value = {
        "service": "udata",
        "data": DOCUMENT_LEGER,
        "meta": {"message_type": "index"},
    }

    for i in range(10):
        print('SEND ONE')
        producer.send("dataset", value=value, key=key)
        producer.flush()


if __name__ == "__main__":
    main()



"""
NB: we can't use pytest-aiohttp helpers beause
it will interfere with the rest of our async code
"""
from datetime import datetime, timedelta
import pytest

from aiohttp.test_utils import TestClient, TestServer

from udata_hydra.app import app_factory

pytestmark = pytest.mark.asyncio


@pytest.fixture
async def client():
    app = await app_factory()
    async with TestClient(TestServer(app)) as client:
        yield client


@pytest.mark.parametrize(
    "query",
    [
        "url=https://example.com/resource-1",
        "resource_id=c4e3a9fb-4415-488e-ba57-d05269b27adf",
    ],
)
async def test_api_latest(setup_catalog, query, client, fake_check):
    await fake_check()
    resp = await client.get(f"/api/checks/latest/?{query}")
    assert resp.status == 200
    data = await resp.json()
    assert data.pop("created_at")
    assert data.pop("id")
    assert data == {
        "response_time": 0.1,
        "deleted": False,
        "resource_id": "c4e3a9fb-4415-488e-ba57-d05269b27adf",
        "catalog_id": 1,
        "domain": "example.com",
        "error": None,
        "url": "https://example.com/resource-1",
        "headers": {"x-do": "you"},
        "timeout": False,
        "dataset_id": "601ddcfc85a59c3a45c2435a",
        "status": 200,
    }


@pytest.mark.parametrize(
    "query",
    [
        "url=https://example.com/resource-1",
        "resource_id=c4e3a9fb-4415-488e-ba57-d05269b27adf",
    ],
)
async def test_api_all(setup_catalog, query, client, fake_check):
    await fake_check(status=500, error="no-can-do")
    await fake_check()
    resp = await client.get(f"/api/checks/all/?{query}")
    assert resp.status == 200
    data = await resp.json()
    assert len(data) == 2
    first, second = data
    assert first["status"] == 200
    assert second["status"] == 500
    assert second["error"] == "no-can-do"


async def test_api_status(setup_catalog, client, fake_check):
    resp = await client.get("/api/status/")
    assert resp.status == 200
    data = await resp.json()
    assert data == {
        "total": 1,
        "pending_checks": 1,
        "fresh_checks": 0,
        "checks_percentage": 0.0,
        "fresh_checks_percentage": 0.0,
    }

    await fake_check()
    resp = await client.get("/api/status/")
    assert resp.status == 200
    data = await resp.json()
    assert data == {
        "total": 1,
        "pending_checks": 0,
        "fresh_checks": 1,
        "checks_percentage": 100.0,
        "fresh_checks_percentage": 100.0,
    }


async def test_api_stats(setup_catalog, client, fake_check):
    resp = await client.get("/api/stats/")
    assert resp.status == 200
    data = await resp.json()
    assert data == {
        "status": [
            {"label": "error", "count": 0, "percentage": 0},
            {"label": "timeout", "count": 0, "percentage": 0},
            {"label": "ok", "count": 0, "percentage": 0},
        ],
        "status_codes": [],
    }

    # only the last one should count
    await fake_check()
    await fake_check(timeout=True, status=None)
    await fake_check(status=500, error="error")
    resp = await client.get("/api/stats/")
    assert resp.status == 200
    data = await resp.json()
    assert data == {
        "status": [
            {"label": "error", "count": 1, "percentage": 100.0},
            {"label": "timeout", "count": 0, "percentage": 0},
            {"label": "ok", "count": 0, "percentage": 0},
        ],
        "status_codes": [{"code": 500, "count": 1, "percentage": 100.0}],
    }


async def test_changed_last_modified(setup_catalog, client, fake_check):
    check = await fake_check(
        headers={
            "last-modified": "Wed, 21 Oct 2015 07:28:00 GMT",
            "content-length": 1,
        }
    )
    resp = await client.get(f"/api/changed/?url={check['url']}")
    assert resp.status == 200
    assert await resp.json() == {
        "changed_at": "2015-10-21T07:28:00",
        "detection": "last-modified",
    }

    # last-modified takes precendence over content-length comparison
    check = await fake_check(
        headers={
            "last-modified": "Wed, 21 Oct 2015 07:28:00 GMT",
            "content-length": 2,
        }
    )
    resp = await client.get(f"/api/changed/?url={check['url']}")
    assert resp.status == 200
    assert await resp.json() == {
        "changed_at": "2015-10-21T07:28:00",
        "detection": "last-modified",
    }


async def test_changed_no_header(setup_catalog, client, fake_check):
    check = await fake_check(headers={})
    resp = await client.get(f"/api/changed/?url={check['url']}")
    assert resp.status == 204


async def test_changed_content_length(setup_catalog, client, fake_check):
    c1 = datetime.now() - timedelta(days=2)
    check = await fake_check(headers={"content-length": 1}, created_at=c1)
    resp = await client.get(f"/api/changed/?url={check['url']}")
    assert resp.status == 204

    c2 = datetime.now() - timedelta(days=1)
    check = await fake_check(headers={"content-length": 2}, created_at=c2)
    resp = await client.get(f"/api/changed/?url={check['url']}")
    assert resp.status == 200
    assert await resp.json() == {
        "changed_at": c2.isoformat(),
        "detection": "content-length",
    }

    c3 = datetime.now()
    check = await fake_check(headers={"content-length": 3}, created_at=c3)
    resp = await client.get(f"/api/changed/?url={check['url']}")
    assert resp.status == 200
    assert await resp.json() == {
        "changed_at": c3.isoformat(),
        "detection": "content-length",
    }

    c4 = datetime.now()
    check = await fake_check(headers={"content-length": 3}, created_at=c4)
    resp = await client.get(f"/api/changed/?url={check['url']}")
    assert resp.status == 200
    assert await resp.json() == {
        "changed_at": c3.isoformat(),
        "detection": "content-length",
    }


async def test_changed_content_length_unchanged(
    setup_catalog, client, fake_check
):
    c1 = datetime.now() - timedelta(days=2)
    check = await fake_check(headers={"content-length": 1}, created_at=c1)
    c2 = datetime.now() - timedelta(days=1)
    check = await fake_check(headers={"content-length": 1}, created_at=c2)
    resp = await client.get(f"/api/changed/?url={check['url']}")
    assert resp.status == 204



from datetime import datetime, timedelta
import json
from unittest.mock import MagicMock
import pytest

import nest_asyncio

from aiohttp.client_exceptions import ClientError
from asyncio.exceptions import TimeoutError
from yarl import URL

from udata_hydra.crawl import crawl, setup_logging


pytestmark = pytest.mark.asyncio
# allows nested async to test async with async :mindblown:
nest_asyncio.apply()


async def test_catalog(setup_catalog, db):
    res = await db.fetch(
        "SELECT * FROM catalog WHERE resource_id = $1",
        "c4e3a9fb-4415-488e-ba57-d05269b27adf",
    )
    assert len(res) == 1
    resource = res[0]
    assert resource["url"] == "https://example.com/resource-1"
    assert resource["dataset_id"] == "601ddcfc85a59c3a45c2435a"


@pytest.mark.parametrize(
    "resource",
    [
        # status, timeout, exception
        (200, False, None),
        (500, False, None),
        (None, False, ClientError("client error")),
        (None, False, AssertionError),
        (None, False, UnicodeError),
        (None, True, TimeoutError),
    ],
)
async def test_crawl(setup_catalog, rmock, event_loop, db, resource, mocker, produce_mock):
    setup_logging()
    status, timeout, exception = resource
    rurl = "https://example.com/resource-1"
    rmock.get(
        rurl,
        status=status,
        headers={"Content-LENGTH": "10", "X-Do": "you"},
        exception=exception,
    )
    event_loop.run_until_complete(crawl(iterations=1))
    assert ("GET", URL(rurl)) in rmock.requests
    res = await db.fetchrow("SELECT * FROM checks WHERE url = $1", rurl)
    assert res["url"] == rurl
    assert res["status"] == status
    if not exception:
        assert json.loads(res["headers"]) == {
            "x-do": "you",
            # added by aioresponses :shrug:
            "content-type": "application/json",
            "content-length": "10",
        }
    assert res["timeout"] == timeout
    if isinstance(exception, ClientError):
        assert res["error"] == "client error"
    elif status == 500:
        assert res["error"] == "Internal Server Error"
    else:
        assert not res["error"]


async def test_backoff(setup_catalog, event_loop, rmock, mocker, fake_check, produce_mock):
    setup_logging()
    await fake_check(resource=2)
    mocker.patch("udata_hydra.config.BACKOFF_NB_REQ", 1)
    mocker.patch("udata_hydra.config.BACKOFF_PERIOD", 0.25)
    magic = MagicMock()
    mocker.patch("udata_hydra.context.monitor").return_value = magic
    rurl = "https://example.com/resource-1"
    rmock.get(rurl, status=200)
    event_loop.run_until_complete(crawl(iterations=1))
    # verify that we actually backed-off
    assert magic.add_backoff.called


async def test_no_backoff_domains(
    setup_catalog, event_loop, rmock, mocker, fake_check, produce_mock
):
    setup_logging()
    await fake_check(resource=2)
    mocker.patch("udata_hydra.config.BACKOFF_NB_REQ", 1)
    mocker.patch("udata_hydra.config.NO_BACKOFF_DOMAINS", ["example.com"])
    magic = MagicMock()
    mocker.patch("udata_hydra.context.monitor").return_value = magic
    rurl = "https://example.com/resource-1"
    rmock.get(rurl, status=200)
    event_loop.run_until_complete(crawl(iterations=1))
    # verify that we actually did not back-off
    assert not magic.add_backoff.called


async def test_excluded_clause(setup_catalog, mocker, event_loop, rmock, produce_mock):
    setup_logging()
    mocker.patch("udata_hydra.config.SLEEP_BETWEEN_BATCHES", 0)
    mocker.patch("udata_hydra.config.EXCLUDED_PATTERNS", ["http%example%"])
    rurl = "https://example.com/resource-1"
    rmock.get(rurl, status=200)
    event_loop.run_until_complete(crawl(iterations=1))
    # url has not been called due to excluded clause
    assert ("GET", URL(rurl)) not in rmock.requests


async def test_outdated_check(setup_catalog, rmock, fake_check, event_loop, produce_mock):
    await fake_check(created_at=datetime.now() - timedelta(weeks=52))
    rurl = "https://example.com/resource-1"
    rmock.get(rurl, status=200)
    event_loop.run_until_complete(crawl(iterations=1))
    # url has been called because check is outdated
    assert ("GET", URL(rurl)) in rmock.requests


async def test_not_outdated_check(
    setup_catalog, rmock, fake_check, event_loop, mocker, produce_mock
):
    mocker.patch("udata_hydra.config.SLEEP_BETWEEN_BATCHES", 0)
    await fake_check()
    rurl = "https://example.com/resource-1"
    rmock.get(rurl, status=200)
    event_loop.run_until_complete(crawl(iterations=1))
    # url has not been called because check is fresh
    assert ("GET", URL(rurl)) not in rmock.requests


async def test_501_get(setup_catalog, event_loop, rmock, produce_mock):
    setup_logging()
    rurl = "https://example.com/resource-1"
    rmock.get(rurl, status=501)
    event_loop.run_until_complete(crawl(iterations=1))
    assert ("GET", URL(rurl)) in rmock.requests



import logging
import csv
import os

from pathlib import Path
from tempfile import NamedTemporaryFile

import aiohttp
import asyncio
import asyncpg
import boto3
from botocore.client import Config
from botocore.exceptions import ClientError
from dotenv import load_dotenv
from humanfriendly import parse_size
from minicli import cli, run, wrap
from progressist import ProgressBar

from udata_event_service.consumer import consume_kafka

from udata_hydra.config import KAFKA_URI
from udata_hydra.kafka.consumer import process_message
from udata_hydra.utils.kafka import get_topic

log = logging.getLogger("udata-hydra")

CATALOG_URL = 'https://www.data.gouv.fr/fr/datasets/r/4babf5f2-6a9c-45b5-9144-ca5eae6a7a6d'

context = {}


@cli
async def init_db(drop=False, table=None, index=False, reindex=False):
    """Create the DB structure"""
    log.info("Initializing database...")
    if drop:
        if table == "catalog" or not table:
            await context["conn"].execute("DROP TABLE IF EXISTS catalog")
        if table == "checks" or not table:
            await context["conn"].execute("DROP TABLE IF EXISTS checks")
    await context["conn"].execute(
        """
        CREATE TABLE IF NOT EXISTS catalog(
            id serial PRIMARY KEY,
            dataset_id VARCHAR(24),
            resource_id UUID,
            url VARCHAR,
            deleted BOOLEAN NOT NULL,
            last_check INT,
            priority BOOLEAN NOT NULL,
            initialization BOOLEAN NOT NULL DEFAULT FALSE,
            UNIQUE(dataset_id, resource_id, url)
        )
    """
    )
    await context["conn"].execute(
        """
        CREATE TABLE IF NOT EXISTS checks(
            id serial PRIMARY KEY,
            url VARCHAR,
            domain VARCHAR,
            created_at TIMESTAMP DEFAULT NOW(),
            status INT,
            headers JSONB,
            timeout BOOLEAN NOT NULL,
            response_time FLOAT,
            error VARCHAR
        )
    """
    )
    if reindex:
        await context["conn"].execute(
            """
            DROP INDEX IF EXISTS url_idx;
            DROP INDEX IF EXISTS domain_idx;
        """
        )
    if index or reindex:
        await context["conn"].execute(
            """
            CREATE INDEX IF NOT EXISTS url_idx ON checks (url);
            CREATE INDEX IF NOT EXISTS domain_idx ON checks (domain);
        """
        )


async def download_file(url, fd):
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as resp:
            while True:
                chunk = await resp.content.read(1024)
                if not chunk:
                    break
                fd.write(chunk)


@cli
async def load_catalog(url=CATALOG_URL):
    """Load the catalog into DB from CSV file"""
    try:
        log.info(f"Downloading catalog from {url}...")
        with NamedTemporaryFile(delete=False) as fd:
            await download_file(url, fd)
        log.info("Upserting catalog in database...")
        # consider everything deleted, deleted will be updated when loading new catalog
        await context["conn"].execute("UPDATE catalog SET deleted = TRUE")
        with open(fd.name) as fd:
            reader = csv.DictReader(fd, delimiter=";")
            rows = list(reader)
            bar = ProgressBar(total=len(rows))
            for row in bar.iter(rows):
                await context["conn"].execute(
                    """
                    INSERT INTO catalog (dataset_id, resource_id, url, deleted, priority, initialization)
                    VALUES ($1, $2, $3, FALSE, FALSE, TRUE)
                    ON CONFLICT (dataset_id, resource_id, url) DO UPDATE SET deleted = FALSE
                """,
                    row["dataset.id"],
                    row["id"],
                    row["url"],
                )
        log.info("Catalog successfully upserted into DB.")
    except Exception as e:
        raise e
    finally:
        fd.close()
        os.unlink(fd.name)


@cli
async def csv_sample(size=1000, download=False, max_size="100M"):
    """Get a csv sample from latest checks

    :size: Size of the sample (how many files to query)
    :download: Download files or just list them
    :max_size: Maximum size for one file (from headers)
    """
    max_size = parse_size(max_size)
    start_q = f"""
        SELECT catalog.resource_id, catalog.dataset_id, checks.url,
            checks.headers->>'content-type' as content_type,
            checks.headers->>'content-length' as content_length
        FROM checks, catalog
        WHERE catalog.last_check = checks.id
        AND checks.headers->>'content-type' LIKE '%csv%'
        AND checks.status >= 200 and checks.status < 400
        AND CAST(checks.headers->>'content-length' AS INTEGER) <= {max_size}
    """
    end_q = f"""
        ORDER BY RANDOM()
        LIMIT {size / 2}
    """
    # get remote stuff for half the sample
    q = f"""{start_q}
        -- ignore ODS, they're correctly formated from a datastore
        AND checks.url NOT LIKE '%/explore/dataset/%'
        AND checks.url NOT LIKE '%/api/datasets/1.0/%'
        -- ignore ours
        AND checks.domain <> 'static.data.gouv.fr'
        {end_q}
    """
    res = await context["conn"].fetch(q)
    # and from us for the rest
    q = f"""{start_q}
        AND checks.domain = 'static.data.gouv.fr'
        {end_q}
    """
    res += await context["conn"].fetch(q)

    data_path = Path("./data")
    dl_path = data_path / "downloaded"
    dl_path.mkdir(exist_ok=True, parents=True)
    if download:
        log.debug("Cleaning up...")
        (data_path / "_index.csv").unlink(missing_ok=True)
        [p.unlink() for p in Path(dl_path).glob("*.csv")]

    lines = []
    bar = ProgressBar(total=len(res))
    for r in bar.iter(res):
        line = dict(r)
        line["resource_id"] = str(line["resource_id"])
        filename = dl_path / f"{r['dataset_id']}_{r['resource_id']}.csv"
        line["filename"] = filename.__str__()
        lines.append(line)
        if not download:
            continue
        await download_file(r["url"], filename.open("wb"))
        with os.popen(f"file {filename} -b --mime-type") as proc:
            line["magic_mime"] = proc.read().lower().strip()
        line["real_size"] = filename.stat().st_size

    with (data_path / "_index.csv").open("w") as ofile:
        writer = csv.DictWriter(ofile, fieldnames=lines[0].keys())
        writer.writeheader()
        writer.writerows(lines)


@cli
def run_kafka_integration() -> None:
    load_dotenv()
    logging.basicConfig(level=logging.INFO)
    # Create bucket if it doesn't exist
    client = boto3.client(
        "s3",
        endpoint_url=os.getenv("MINIO_URL"),
        aws_access_key_id=os.getenv("MINIO_USER"),
        aws_secret_access_key=os.getenv("MINIO_PWD"),
        config=Config(signature_version="s3v4"),
    )
    try:
        client.head_bucket(Bucket=os.getenv("MINIO_BUCKET"))
    except ClientError:
        client.create_bucket(Bucket=os.getenv("MINIO_BUCKET"))

    def run_process_message(key: str, data: dict, topic: str) -> None:
        try:
            asyncio.get_event_loop().run_until_complete(process_message(key, data, topic))
        except Exception as e:
            log.error(e)

    consume_kafka(
        kafka_uri=KAFKA_URI,
        group_id="datalake",
        topics=[get_topic(topic_suffix) for topic_suffix in ["resource.created", "resource.modified", "resource.deleted"]],
        message_processing_func=run_process_message,
    )


@wrap
async def cli_wrapper():
    dsn = os.getenv(
        "DATABASE_URL", "postgres://postgres:postgres@localhost:5432/postgres"
    )
    context["conn"] = await asyncpg.connect(dsn=dsn)
    context["dsn"] = dsn
    yield
    await context["conn"].close()


if __name__ == "__main__":
    run()
